/* vim: set shiftwidth=4 tabstop=8 softtabstop=4: */
/* $Id: libspt.c,v 1.19 2002/12/13 18:53:44 shinra Exp $ */

#define IN_LIBSPT
#include "sptprivate.h"
#include <sys/wait.h>
#include <time.h>
#include <stdio.h>  /* for perror() */
#include <signal.h>
#ifdef HAVE_PTY_SVR4
#include <sys/stropts.h>
#endif
#ifdef HAVE_SYS_IOCTL_H /* for TIOCSCTTY */
#include <sys/ioctl.h>
#endif

struct spt_handle_tag {
    connection *pconn;
    char *ttyname;    /* slave */
    unsigned int ptytype;
    int masterfd;
    int agentfd;
    int reply;		    /* refered outside wait_for_reply */
    int msgprocerr;	    /* used only in wait_for_reply */
    pid_t agentpid;
    dispatch_table *client_tab;
    spt_cleanup_proc cleanup;
    void *cleanuparg;
    char *utmphost;
    pid_t utmppid;
    int exit_st;
#ifdef HAVE_PTY_SVR4
    int grantpt_failed;
#endif
};

typedef struct {
    int type;
    int (*openproc) (spt_handle *phandle);
} ptyopen_table_item;

#ifdef HAVE_PTY_SVR4
static int open_pty_svr4(spt_handle *phandle);
#endif
#ifdef HAVE_PTY_BSD
static int open_pty_bsd(spt_handle *phandle);
#endif
#ifdef HAVE_PTY_PTC
static int open_pty_ptc(spt_handle *phandle);
#endif
#ifdef HAVE_PTY_CLONE
static int open_pty_clone(spt_handle *phandle);
#endif
#ifdef HAVE_PTY_SGI4
static int open_pty_sgi4(spt_handle *phandle);
#endif
#ifdef HAVE_PTY_NUMERIC
static int open_pty_numeric(spt_handle *phandle);
#endif

static int wait_for_reply(spt_handle *phandle);
static int send_and_wait(spt_handle *phandle, const msgunion_with_type *pmsg);
static int try_dup2(int oldfd, int newfd);
static void shutdown_agent(pid_t pid);
static int waitpid_with_timeout(pid_t pid, const struct timespec *ptimeout);
static void sigchld_handler(int dummy);
static void disconnect_before_close(spt_handle *phandle);
static int start_agent(spt_handle *phandle);
static int trivialcmd(spt_handle *phandle, const msgunion_with_type *pmsg,
	int errreply, int mappederr);
static MSGPROC(msg_result_proc);
static MSGPROC(msg_proto_violation_proc);

const dispatch_table_item client_tab_items[] = {
    { MSG_RESULT, &msg_result_proc },
    { MSG_PROTO_VIOLATION, &msg_proto_violation_proc },
};

char *const agent_args[] = {
    AGENT_BASENAME,
    NULL
};

static ptyopen_table_item ptyopen_tab[] = {
#ifdef HAVE_PTY_PTC
    { PTYTYPE_PTC, &open_pty_ptc },
#endif
#ifdef HAVE_PTY_CLONE
    { PTYTYPE_CLONE, &open_pty_clone },
#endif
#ifdef HAVE_PTY_SGI4
    { PTYTYPE_SGI4, &open_pty_sgi4 },
#endif
#ifdef HAVE_PTY_NUMERIC
    { PTYTYPE_NUMERIC, &open_pty_numeric },
#endif
#ifdef HAVE_PTY_BSD
    { PTYTYPE_BSD, &open_pty_bsd },
#endif
#ifdef HAVE_PTY_SVR4
    { PTYTYPE_SVR4, &open_pty_svr4 },
#endif
#ifdef HAVE_PTY_CYGWIN
    { PTYTYPE_CYGWIN, &open_pty_cygwin },
#endif
};

int SPT_EXPORT
spt_open_pty(spt_handle **pphandle, int *pmasterfd,
	spt_cleanup_proc cleanup, void *cleanuparg)
{
    spt_handle *phandle;
    dispatch_table *client_tab;
    int retval;
    unsigned int i;
    msgunion_with_type msg;
    
    if ((phandle = MALLOC(sizeof(spt_handle))) == NULL)
	return SPT_E_NOMEM;
    phandle->cleanup = cleanup;
    phandle->cleanuparg = cleanuparg;
    phandle->utmppid = getpid();
    phandle->exit_st = 0;

    retval = SPT_E_NOMEM;
    if ((phandle->utmphost = STRDUP("(unknown)")) == NULL)
	goto err1;

    if ((client_tab = MALLOC(sizeof(dispatch_table))) == NULL)
	goto err2;

    client_tab->items = client_tab_items;
    client_tab->n_items = ARRAYLEN(client_tab_items);
    client_tab->defproc = NULL;
    client_tab->arg = phandle;
    phandle->client_tab = client_tab;
    for (i = 0; i < ARRAYLEN(ptyopen_tab); ++i) {
	phandle->ptytype = ptyopen_tab[i].type;
	retval = (*ptyopen_tab[i].openproc) (phandle);
	if (retval)
	    continue;
	retval = start_agent(phandle);
	if (!retval)
	    goto agent_ok;
	FREE(phandle->ttyname);
	close(phandle->masterfd);
    }
    goto ptyfail;
agent_ok:
#ifdef HAVE_PTY_SVR4
    if (phandle->ptytype == PTYTYPE_SVR4) {
	retval = phandle->grantpt_failed ? SPT_E_CHOWN_FAIL : SPT_E_NONE;
	goto success;
    }
#endif
#ifdef HAVE_PTY_CYGWIN
    if (phandle->ptytype == PTYTYPE_CYGWIN)
	goto success;
#endif
#ifdef HAVE_PTY_SGI4
    if (phandle->ptytype == PTYTYPE_SGI4)
	goto success;
#endif
    msg.msgtype = MSG_INIT_TTY;
    retval = send_and_wait(phandle, &msg);
    if (retval == SPT_E_NONE) {
	if (phandle->reply == ERR_CHOWN_FAIL)
	    retval = SPT_E_CHOWN_FAIL;
	else if (phandle->reply != ERR_NONE)
	    retval = SPT_E_AGENT_FATAL;
    }
    if (retval == SPT_E_NONE || retval == SPT_E_CHOWN_FAIL)
	goto success;

    if (retval != SPT_E_AGENT_DEAD && retval != SPT_E_AGENT_FATAL) {
	msg.msgtype = MSG_DISCONNECT;
	send_and_wait(phandle, &msg);
    }
    disconnect_before_close(phandle);
    close(phandle->masterfd);
    FREE(phandle->ttyname);
    goto ptyfail;

success:
#if HAVE_REVOKE
    /* XXX: silently fail */
    revoke(phandle->ttyname);
#endif
    *pphandle = phandle;
    *pmasterfd = phandle->masterfd;
    return retval;

ptyfail:
    FREE(client_tab);
err2:
    FREE(phandle->utmphost);
err1:
    FREE(phandle);
    return retval;
}

int SPT_EXPORT
spt_close_pty(spt_handle *phandle)
{
    int r;
    int retval = SPT_E_CLEANUP_FAIL;
    connection *pconn = phandle->pconn;

    if (pconn) {
	msgunion_with_type msg;
	msg.msgtype = MSG_DISCONNECT;
	r = send_and_wait(phandle, &msg);
	if (!r && phandle->reply == ERR_NONE)
	    retval = SPT_E_NONE;
	disconnect_before_close(phandle);
    }
    close(phandle->masterfd);
    FREE(phandle->ttyname);
    FREE((dispatch_table *)phandle->client_tab);
    FREE(phandle->utmphost);
    FREE(phandle);
    return retval;
}

/*
 * If received MSG_RESULT, returns SPT_E_NONE and the result is stored
 * into phandle->reply. Otherwise, returns an appropreate error.
 * None of SPT_IE_* is returned.
 */
static int
wait_for_reply(spt_handle *phandle)
{
    int retval;
    connection *pconn = phandle->pconn;

    while ((retval = spt_p_do_read(pconn)) == SPT_IE_AGAIN);
    if (retval == SPT_E_NONE)
	retval = phandle->msgprocerr;
    if (retval == SPT_IE_PEERDEAD)
	return SPT_E_AGENT_DEAD;
    else if (retval == SPT_IE_VIOLATION)
	return SPT_E_AGENT_FATAL;
    else
	return retval;
}

static int
send_and_wait(spt_handle *phandle, const msgunion_with_type *pmsg)
{
    int retval;

    DEBUG(fprintf(stderr, "send_and_wait: type=%d\n", pmsg->msgtype));
    retval = spt_p_send_msg(phandle->pconn, pmsg);
    if (retval) {
	if (retval == SPT_IE_PEERDEAD)
	    retval = SPT_E_AGENT_DEAD;
	return retval;
    }
    return wait_for_reply(phandle);
}

static int
try_dup2(int oldfd, int newfd)
{
    int r;
    r = dup2(oldfd, newfd);
    if (r < 0)
	_exit(1);
    return r;
}

static int
start_agent(spt_handle *phandle)
{
    int r;
    int fdpair[2];
    pid_t pid;
    int retval;

#ifdef SPT_USE_SOCKET
    r = socketpair(PF_UNIX, SOCK_STREAM, 0, fdpair);
#else
    /* pipes must support bidirectional read/write */
    r = pipe(fdpair);
#endif
    if (r) {
	close(phandle->masterfd);
	FREE(phandle->ttyname);
	return SPT_E_RESOURCE;
    }
    retval = spt_p_create_conn(
	    &phandle->pconn, phandle->client_tab, fdpair[0]);
    if (retval)
	goto err1;

    pid = fork();
    if (pid < 0) {
	retval = SPT_E_RESOURCE;
	goto err2;
    } else if (pid == 0) {
	/* child */
	if (phandle->cleanup)
	    (*phandle->cleanup) (phandle->cleanuparg);  /* must do first */
	close(fdpair[0]);
	if (phandle->masterfd == AGENT_CONN_FD) {
	    if (fdpair[1] == AGENT_MASTER_FD) {
		int tmpfd;
		tmpfd = dup(AGENT_CONN_FD);
		if (tmpfd < 0)
		    _exit(1);
		try_dup2(AGENT_MASTER_FD, AGENT_CONN_FD);
		try_dup2(tmpfd, AGENT_MASTER_FD);
		close(tmpfd);
	    } else {
		/* always fdpair[1] != AGENT_CONN_FD */
		try_dup2(AGENT_CONN_FD, AGENT_MASTER_FD);
		try_dup2(fdpair[1], AGENT_CONN_FD);
		close(fdpair[1]);
	    }
	} else if (fdpair[1] == AGENT_MASTER_FD) {
	    /* always phandle->masterfd != AGENT_MASTER_FD */
	    try_dup2(AGENT_MASTER_FD, AGENT_CONN_FD);
	    try_dup2(phandle->masterfd, AGENT_MASTER_FD);
	    close(phandle->masterfd);
	} else {
	    if (phandle->masterfd != AGENT_MASTER_FD) {
		try_dup2(phandle->masterfd, AGENT_MASTER_FD);
		close(phandle->masterfd);
	    }
	    if (fdpair[1] != AGENT_CONN_FD) {
		try_dup2(fdpair[1], AGENT_CONN_FD);
		close(fdpair[1]);
	    }
	}
	execve(AGENT_PATH, agent_args, NULL);
	/* exec failure */
	perror("cannot execute sptagent");
	_exit(1);
	/* NOTREACHED */
    } else {
	/* parent */
	msgunion_with_type msg;

	close(fdpair[1]);
	msg.msgtype = MSG_CONNECT;
	msg.mu.mu_connect.version = SPT_PROTO_VERSION;
	msg.mu.mu_connect.ttyname = phandle->ttyname;
	msg.mu.mu_connect.ptytype = phandle->ptytype;
	retval = send_and_wait(phandle, &msg);
	/*
	 * If the connection was closed before first reply to MSG_CONNECT, it
	 * was probably because exec() failed. But it also may be because
	 * the agent terminated abnormally, or something other than sptagent
	 * was placed in AGENT_PATH and it closed the pipe. So, we must check
	 * whether my child is alive or not, and call wait() properly.
	 * Anyway, we return SPT_E_AGENT_EXEC, which is the best assumption.
	 */
	if (retval) {
	    if (retval == SPT_IE_PEERDEAD)
		retval = SPT_E_AGENT_EXEC;
	    goto err2;
	}
	if (phandle->reply == ERR_VERSION_MISMATCH
		|| phandle->reply == ERR_UNKNOWN_PTYTYPE)
	    retval = SPT_E_AGENT_VERSION;
	else if (phandle->reply == ERR_WRONG_INST)
	    retval = SPT_E_AGENT_INST;
	else if (phandle->reply != ERR_NONE)
	    retval = SPT_E_AGENT_FATAL;
	if (retval)
	    goto err2;

	phandle->agentfd = fdpair[0];
	phandle->agentpid = pid;
	return SPT_E_NONE;
err2:
	shutdown_agent(pid);
    }
err1:
    close(fdpair[0]);
    return retval;
}

static void
shutdown_agent(pid_t pid)
{
    struct timespec timeout;

    timeout.tv_sec = 0;
    timeout.tv_nsec = 100 * 1000 * 1000; 
    if (waitpid_with_timeout(pid, &timeout)) {
	kill(pid, SIGTERM);
	if (waitpid_with_timeout(pid, &timeout)) {
	    kill(pid, SIGKILL);
	    timeout.tv_sec = 2;
	    timeout.tv_nsec = 0;
	    waitpid_with_timeout(pid, &timeout);
	    /* return anyway */
	}
    }
}

/* FIXME: sigaction */
static int
waitpid_with_timeout(pid_t pid, const struct timespec *ptimeout)
{
    struct timespec work;
    static void (*old_sigchld) (int dummy);

    if (waitpid(pid, NULL, WNOHANG) == pid)
	return 0;
    old_sigchld = signal(SIGCHLD, &sigchld_handler);
    work = *ptimeout;
    while (nanosleep(&work, &work) && errno == EINTR) {
	if (waitpid(pid, NULL, WNOHANG) == pid) {
	    signal(SIGCHLD, old_sigchld);
	    return 0;
	}
    }
    signal(SIGCHLD, old_sigchld);
    return 1;
}

static void
sigchld_handler(int dummy)
{
    /* do nothing */
}

void SPT_EXPORT
spt_detach_handle(spt_handle *phandle)
{
    if (phandle->pconn) {
	spt_p_destroy_conn(phandle->pconn);
	close(phandle->agentfd);
    }
    FREE(phandle->ttyname);
    FREE((dispatch_table *)phandle->client_tab);
    FREE(phandle);
}

int SPT_EXPORT
spt_open_slave(const spt_handle *phandle)
{
    return open(phandle->ttyname, O_RDWR | O_NOCTTY);
}

int SPT_EXPORT
spt_init_slavefd(const spt_handle *phandle, int slavefd)
{
    /* XXX: screen requires !linux, !sgi, !__osf__ and !M_UNIX */
#if defined(HAVE_PTY_SVR4) && defined(I_PUSH)
#if 0	/* only rxvt and mlterm do this */
#ifdef HAVE_ISASTREAM
    if (isastram(slavefd))
#endif
#endif
    {
	if (ioctl(slavefd, I_PUSH, "ptem")
		/* XXX: "consem" fails on solaris 8. */
#if !defined(SVR4) && !(defined(SYSV) && defined(i386)) && 0
		|| (!getenv("CONSEM") && ioctl(slavefd, I_PUSH, "consem"))
#endif
		|| ioctl(slavefd, I_PUSH, "ldterm")
#ifndef __hpux
		|| ioctl(slavefd, I_PUSH, "ttcompat")
#endif
		)
	    return SPT_E_TTYINIT_FAIL;
    }
#endif /* HAVE_PTY_SVR4 && I_PUSH */
    return SPT_E_NONE;
}

int SPT_EXPORT
spt_get_slavename(const spt_handle *phandle, char **pname)
{
    char *res = STRDUP(phandle->ttyname);
    if (res == NULL)
	return SPT_E_NOMEM;
    *pname = res;
    return SPT_E_NONE;
}

void SPT_EXPORT
spt_free_obj(void *pobj)
{
    SAFE_FREE(pobj);
}

int SPT_EXPORT
spt_utmp_set_host(spt_handle *phandle, const char *host)
{
    char *newhost = STRDUP(host);
    if (newhost == NULL)
	return SPT_E_NOMEM;
    FREE(phandle->utmphost);
    phandle->utmphost = newhost;
    return SPT_E_NONE;
}

int SPT_EXPORT
spt_utmp_set_pid(spt_handle *phandle, pid_t pid)
{
    phandle->utmppid = pid;
    return SPT_E_NONE;
}

int SPT_EXPORT
spt_utmp_set_exit(spt_handle *phandle, int exit_st)
{
    phandle->exit_st = exit_st;
    return SPT_E_NONE;
}

static int
trivialcmd(spt_handle *phandle, const msgunion_with_type *pmsg,
	int errreply, int mappederr)
{
    int retval;

    if (phandle->pconn == NULL)
	return mappederr;
    retval = send_and_wait(phandle, pmsg);
    if (retval == SPT_E_NONE) {
	if (phandle->reply == errreply)
	    retval = mappederr;
	else if (phandle->reply != ERR_NONE)
	    retval = SPT_E_AGENT_FATAL;
    }
    if (retval == SPT_E_AGENT_DEAD || retval == SPT_E_AGENT_FATAL)
	disconnect_before_close(phandle);
    return retval;
}

int SPT_EXPORT
spt_login_utmp(spt_handle *phandle)
{
    msgunion_with_type msg;
    msg.msgtype = MSG_LOGIN_UTMP;
    msg.mu.mu_utmp.host = phandle->utmphost;
    msg.mu.mu_utmp.pid = (long)phandle->utmppid;
    return trivialcmd(phandle, &msg, ERR_UTMP_FAIL, SPT_E_UTMP_FAIL);
}

int SPT_EXPORT
spt_logout_utmp(spt_handle *phandle)
{
    msgunion_with_type msg;
    msg.msgtype = MSG_LOGOUT_UTMP;
    msg.mu.mu_int = phandle->exit_st;
    return trivialcmd(phandle, &msg, ERR_UTMP_FAIL, SPT_E_UTMP_FAIL);
}

int SPT_EXPORT
spt_login_wtmp(spt_handle *phandle)
{
    msgunion_with_type msg;
    msg.msgtype = MSG_LOGIN_WTMP;
    msg.mu.mu_utmp.host = phandle->utmphost;
    msg.mu.mu_utmp.pid = (long)phandle->utmppid;
    return trivialcmd(phandle, &msg, ERR_UTMP_FAIL, SPT_E_UTMP_FAIL);
}

int SPT_EXPORT
spt_logout_wtmp(spt_handle *phandle)
{
    msgunion_with_type msg;
    msg.msgtype = MSG_LOGOUT_WTMP;
    msg.mu.mu_int = phandle->exit_st;
    return trivialcmd(phandle, &msg, ERR_UTMP_FAIL, SPT_E_UTMP_FAIL);
}

int SPT_EXPORT
spt_update_lastlog(spt_handle *phandle)
{
    msgunion_with_type msg;
    msg.msgtype = MSG_UPDATE_LASTLOG;
    msg.mu.mu_utmp.host = phandle->utmphost;
    msg.mu.mu_utmp.pid = (long)phandle->utmppid;
    return trivialcmd(phandle, &msg, ERR_UTMP_FAIL, SPT_E_UTMP_FAIL);
}

int SPT_EXPORT
spt_set_ctty(int slavefd)
{
#ifdef TIOCSCTTY
    return ioctl(slavefd, TIOCSCTTY, NULL);
#else
    int fd;
    const char *name;
    fd = open("/dev/tty", O_RDWR | O_NOCTTY);
    if (fd >= 0) {
	/* already has controlling tty */
	close(fd);
	return -1;
    }
    name = ttyname(slavefd);
    if (!name)
	return -1;
    fd = open(name, O_RDWR);
    if (fd < 0)
	return -1;
    close(fd);
    fd = open("/dev/tty", O_RDWR | O_NOCTTY);
    if (fd < 0)	/* does not have controlling tty yet */
	return -1;
    close(fd);
    return 0;
#endif
}

static void
disconnect_before_close(spt_handle *phandle)
{
    assert(phandle->pconn);
    spt_p_destroy_conn(phandle->pconn);
    phandle->pconn = NULL;
    close(phandle->agentfd);
    shutdown_agent(phandle->agentpid);
}

static MSGPROC(msg_result_proc) {
    spt_handle *phandle = arg;
    assert(msgtype == MSG_RESULT);
    phandle->msgprocerr = SPT_E_NONE;
    phandle->reply = pmu->mu_int;
    return SPT_E_NONE;
}

static MSGPROC(msg_proto_violation_proc) {
    spt_handle *phandle = arg;
    assert(msgtype == MSG_PROTO_VIOLATION);
    phandle->msgprocerr = SPT_E_AGENT_FATAL;
    return SPT_E_NONE;
}

/* tested */
#ifdef HAVE_PTY_SVR4
static int
open_pty_svr4(spt_handle *phandle)
{
    int retval;
    int masterfd;
    const char *ptytmp;
    static void (*old_sigchld) (int dummy);

    masterfd = phandle->masterfd = open("/dev/ptmx", O_RDWR | O_NOCTTY);
    if (masterfd < 0)
	return SPT_E_NOTTY;
    old_sigchld = signal(SIGCHLD, &sigchld_handler);
    phandle->grantpt_failed = grantpt(masterfd);
    signal(SIGCHLD, old_sigchld);
    if (unlockpt(masterfd)) {
	retval = SPT_E_NOTTY;
	goto err1;
    }
    if ((ptytmp = ptsname(masterfd)) == NULL) {
	retval = SPT_E_NOTTY;
	goto err1;
    }
    if ((phandle->ttyname = STRDUP(ptytmp)) == NULL) {
	retval = SPT_E_NOMEM;
	goto err1;
    }
    return SPT_E_NONE;
err1:
    close(masterfd);
    return retval;
}
#endif /* HAVE_PTY_SVR4 */

/* tested */
#ifdef HAVE_PTY_BSD
static int
open_pty_bsd(spt_handle *phandle)
{
    int masterfd;
    char *name;
    const char *p1, *p2;

    if ((name = STRDUP("/dev/pty??")) == NULL)
	return SPT_E_NOMEM;
    for (p1 = "pqrstuvwxyzPQRST"; *p1; ++p1) {
	name[8] = *p1;
	for (p2 = "0123456789abcdef"; *p2; ++p2) {
	    name[9] = *p2;
	    masterfd = open(name, O_RDWR | O_NOCTTY);
	    if (masterfd < 0) {
		if (errno == EIO)
		    continue;
		goto notty;
	    }
	    name[5] = 't';
	    phandle->masterfd = masterfd;
	    phandle->ttyname = name;
	    return SPT_E_NONE;
	}
    }
notty:
    FREE(name);
    return SPT_E_NOTTY;
}
#endif /* HAVE_PTY_BSD */

/* UNTESTED */
#ifdef HAVE_PTY_PTC
/* AIX, SGI3 */
static int
open_pty_ptc_aix(spt_handle *phandle)
{
    int retval;
    int masterfd;
    const char *ptytmp;

    masterfd = phandle->masterfd = open("/dev/ptc", O_RDWR | O_NOCTTY);
    if (masterfd < 0)
	return SPT_E_NOTTY;
    if ((ptytmp = ttyname(masterfd)) == NULL) {
	retval = SPT_E_NOTTY;
	goto err1;
    }
    if ((phandle->ttyname = STRDUP(ptytmp)) == NULL) {
	retval = SPT_E_NOMEM;
	goto err1;
    }
    return SPT_E_NONE;
err1:
    close(masterfd);
    return retval;
}
#endif /* HAVE_PTY_PTC */

/* UNTESTED */
#ifdef HAVE_PTY_CLONE
/* HPUX */
static int
open_pty_clone(spt_handle *phandle)
{
    int retval;
    int masterfd;
    const char *ptytmp;

    masterfd = phandle->masterfd = open("/dev/ptym/clone", O_RDWR | O_NOCTTY);
    if (masterfd < 0)
	return SPT_E_NOTTY;
    if ((ptytmp = ptsname(masterfd)) == NULL) {
	retval = SPT_E_NOTTY;
	goto err1;
    }
    if ((phandle->ttyname = STRDUP(ptytmp)) == NULL) {
	retval = SPT_E_NOMEM;
	goto err1;
    }
    return SPT_E_NONE;
err1:
    close(masterfd);
    return retval;
}
#endif /* HAVE_PTY_CLONE */

/* UNTESTED */
#ifdef HAVE_PTY_SGI4
/* SGI4 _getpty() */
static int
open_pty_sgi4(spt_handle *phandle)
{
    int retval;
    int masterfd;
    const char *ptytmp;
    static void (*old_sigchld) (int dummy);

    old_sigchld = signal(SIGCHLD, &sigchld_handler);
    ptytmp = _getpty(&masterfd, O_RDWR | O_NOCTTY | O_NDELAY, 0622, 0);
    signal(SIGCHLD, old_sigchld);
    if (masterfd < 0)
	return SPT_E_NOTTY;
    if ((phandle->ttyname = STRDUP(ptytmp)) == NULL) {
	retval = SPT_E_NOMEM;
	goto err1;
    }
    phandle->masterfd = masterfd;
    return SPT_E_NONE;
err1:
    close(masterfd);
    return retval;
}
#endif /* HAVE_PTY_SGI4 */

/* UNTESTED */
#ifdef HAVE_PTY_NUMERIC
/* SCO */
static int
open_pty_numeric(spt_handle *phandle)
{
    int masterfd;
    char *name;
    int i;

    if ((name = STRDUP("/dev/ptyp???")) == NULL)
	return SPT_E_NOMEM;
    for (i = 0; i < 256; ++i) {
	sprintf(name + 9, "%d", i);
	masterfd = open(name, O_RDWR | O_NOCTTY);
	if (masterfd < 0) {
	    if (errno == EIO)
		continue;
	    goto notty;
	}
	name[5] = 't';
	phandle->masterfd = masterfd;
	phandle->ttyname = name;
	return SPT_E_NONE;
    }
notty:
    FREE(name);
    return SPT_E_NOTTY;
}
#endif /* HAVE_PTY_NUMERIC */

/* UNTESTED */
#ifdef HAVE_PTY_CYGWIN
static int
open_pty_cygwin(spt_handle *phandle)
{
    int retval;
    int masterfd;
    const char *ptytmp;
    static void (*old_sigchld) (int dummy);

    masterfd = phandle->masterfd = open("/dev/ptmx", O_RDWR | O_NOCTTY);
    if (masterfd < 0)
	return SPT_E_NOTTY;
    if ((ptytmp = ptsname(masterfd)) == NULL) {
	retval = SPT_E_NOTTY;
	goto err1;
    }
    if ((phandle->ttyname = STRDUP(ptytmp)) == NULL) {
	retval = SPT_E_NOMEM;
	goto err1;
    }
    return SPT_E_NONE;
err1:
    close(masterfd);
    return retval;
}
#endif /* HAVE_PTY_CYGWIN */

